const PccApp = require('../../../picockpit-library/picockpit-base-app')
const { Chip, Line } = require("node-libgpiod");
let Pigpio


class GpioApp extends PccApp{
    constructor(){
        super("gpio")
        this.logger.info("Starting gpio")
        
        this.raspiVersion = 5
        
        if (this.raspiVersion === 5){
            this.chip = new Chip("gpiochip0")
        }
        else {
            Pigpio = require('pigpio').Gpio
        }

        this.managedPins = Array(28).fill(0)
        this.inputPins = Array(28).fill(0)
        this.outputPins = Array(28).fill(0)
        this.pwmPins = Array(28).fill(0)

        this.client.on("message", (topic, message) => {
            message = message.length !== 0 ? JSON.parse(message) : message
            this.onControlMessage(topic, message)
        })

        this.readPins()
    }

    onControlMessage = (topic, message) => {
        // New pin - topic is set/<pin_id>/setup
            // Input pin - message.data.target === 0
            // Output pin - message.data.target === 1
        
        // Change pin state - set/<pin_id>
            // message.state --> desired pin state

        // Delete pin - topic is set/<pin_id>/setup
            // message is empty buffer 

        // console.log("---- Recieved Message ----")

        /* this.logger.warn(topic)
        this.logger.warn(message) */

        if (!topic.includes("gpio/set/")){
            this.logger.debug("unhandled message")
            return
        }

        // Delete pin
        else if (message.length === 0){
            this.logger.debug("Deleting pin?")
            const splitTopic = topic.split("/")
            const deleting = splitTopic[splitTopic.length-1] === "setup" ? true : false  
            
            if (deleting){
                const pinId = parseInt(splitTopic[splitTopic.length-2])
                this.logger.debug("DELETING PIN")
                this.logger.debug(`pinId: ${pinId}`)
    
                this.removePin(pinId)
            }
        }

        // New pin
        else if (message.hasOwnProperty("data")){
            this.logger.debug("NEW PIN")
            const pinId = message.data.bcmId
            this.logger.debug(`pinId: ${pinId}`)

            this.setUpPin(pinId, message.data.target)
        }

        // Set pin
        else if (message.hasOwnProperty("state")){
            this.logger.debug("SETTING PIN")
            const splitTopic = topic.split("/")
            const pinId = parseInt(splitTopic[splitTopic.length-1])
            this.logger.debug(`pinId: ${pinId}`)

            this.setPinState(pinId, message.state)
        }

        // Set pwm pin

        else if (message.hasOwnProperty("freq")){
            this.logger.debug("SETTING UP PWM PIN")
            const splitTopic = topic.split("/")
            const pinId = parseInt(splitTopic[splitTopic.length-1])
            this.logger.debug(`pinId: ${pinId}`)

            if (this.raspiVersion !== 5){
                this.setUpPin(pinId, 2)
                this.setPwmPin(message.freq, message.dutyCycle)
            }

        }

        else {
            this.logger.debug(topic)
            this.logger.debug(message)
        }
    }

    setUpPin = (id, pinType) => {
        /* this.logger.warn(`id: ${id}, pinType: ${pinType}`)
        this.logger.warn(this.managedPins)
        this.logger.warn(this.inputPins)
        this.logger.warn(this.outputPins) */

        this.managedPins[id] = pinType + 1 // +1 so zero stands for not managed --> 1 for input, 2 for output bin, 3 for PWM

        if (pinType === 0){
            if (this.raspiVersion === 5){
                this.inputPins[id] = new Line(this.chip, id)
                this.inputPins[id].requestInputMode()
            }
            else {
                this.inputPins[id] = new Pigpio(id, {mode: Pigpio.INPUT})
            }
        }

        if (pinType === 1){
            if (this.raspiVersion === 5){
                this.outputPins[id] = new Line(this.chip, id)
                this.outputPins[id].requestOutputMode()
            }
            else {
                this.outputPins[id] = new Pigpio(id, {mode: Pigpio.OUTPUT})
            }
        }

        if(pinType === 2){
            if (this.raspiVersion !== 5){
                this.pwmPins[id] = new Pigpio(pinId, {mode: Pigpio.OUTPUT})
            }
        }
    }

    setPinState = (id, state) => {
        if (this.raspiVersion === 5){
            this.outputPins[id].setValue(state ? 1 : 0)
        }
        else {
            this.outputPins[id].digitalWrite(state)
        }
        this.publish(`$state/${id}`, {"state": state ? 1 : 0}, {retain: true})
    }

    setPwmPin = (id, freq, dutyCycle) => {
        this.pwmPins[id].pwmRange(100)
        this.pwmPins[id].pwmFrequency(freq)
        this.pwmPins[id].pwmWrite(dutyCycle)
    }

    removePin = (id) => {
        /* this.logger.warn(`id: ${id}`)
        this.logger.warn(this.managedPins)
        this.logger.warn(this.inputPins)
        this.logger.warn(this.outputPins) */

        const pinType = this.managedPins[id] - 1 // see setup pin why -1
        this.managedPins[id] = 0

        if (pinType === 0){
            if (this.raspiVersion === 5){
                this.inputPins[id].release()
            }
            this.inputPins[id] = 0
        }
        else {
            if (this.raspiVersion === 5){
                this.outputPins[id].release()
            }
            this.outputPins[id] = 0
        }
    }

    readPins = async () => {
        while (true){
            await this.sleep(1000)
            for (let i = 0; i < this.managedPins.length; i++){
                if (this.managedPins[i] === 1){
                    let value
                    if (this.raspiVersion === 5){
                        value = this.inputPins[i].getValue()
                    }
                    else {
                        value = this.inputPins[i].digitalRead() 
                    }
                    this.publish(`$state/${i}`, {"state": value}, {retain:true})
                }
            }
        }  
    }

    sleep = (millis) => {
        // console.log(`Sleeping for ${millis} ms`)
        return new Promise(resolve => setTimeout(resolve, millis));
    }
}

gpio = new GpioApp

/* const PccApp = require('../../../picockpit-library/picockpit-base-app')
const { spawnSync } = require('child_process')
const Pigpio = require('pigpio').Gpio

class GpioApp extends PccApp {
    constructor(){
        super("gpio")
        //this.debugLog("Constructing gpio")

        this.managedPins = Array(28).fill(0)
        this.inPins = [] // mode 1
        this.outPinsOnOff = Array(28).fill(false) // mode 2
        this.outPinsSoftwarePWM = Array(28).fill(false) // mode 3

        this.client.on('message', (topic, message) => {
            if (message.length > 0){
                message = JSON.parse(message)
            }
            //this.debugLog(`Received message, topic: ${topic}, message: ${message}`)
            this.onControlMessage(topic, message)    
        })

        this.initialI2cState()

        this.readPins()
    }

    initialI2cState = async () => {
        await this.sleep(5000)
        
        const publish_i2c_state = {
            "i2cAvailable": true,
            "i2cActive": !this.readI2c()
        }
        this.publish("resources/$state/i2c", publish_i2c_state, {appName: "core", retain: true})
    }

    onControlMessage = async (topic, message) => {
        if (topic.includes("/gpio/set/") && !topic.includes("defaultpull") && !topic.includes("i2c")){
            const temp = topic.substring(topic.length - 10, topic.length);
            const pinId = parseInt(temp.match(/\d+/)[0])
            if (topic.includes("setup") && message.length !== 0){
                // New pin or reconfig
                //this.debugLog('New pin or reconfiguring pin')
                this.newOrReconfigPin(message)
            }
            else if (!topic.includes("setup") && message.length !== 0){
                // Output pin state changed
                //this.debugLog('Changing state of output pin')
                if (!this.managedPins[pinId]){
                    await this.sleep(1000)
                }
                this.changePinState(pinId, message.state, message)
            }
            else if (topic.includes("setup") && message.length === 0){
                // Pin removed
                //this.debugLog('Removing pin')
                this.stopManagingPin(pinId)
            }
        }
        else if (topic.includes("gpio/$state/")){
            // this.verboseLog(`Topic - state: ${topic}`)
        }
        else if (topic.includes("core/resources/set/")){
            // this.verboseLog(`Topic - core/ressources: ${topic}`)
            if(topic.includes("i2c")){
                //this.debugLog("Setting I2c")
                this.setI2c(message)
            }
        }
        else {
            // this.verboseLog(`Topic - other: ${topic}`)
        }
    }

    newOrReconfigPin = (message) => {
        // this.verboseLog('pin content', message)
        const pinId = message.data.bcmId
        this.stopManagingPin(pinId)
        this.initializePin(pinId, message.data.target + 1, message)
        //this.debugLog('Done reconfiguring / setting up pin')
    }

    changePinState = (pinId, state, message) => {
        if (message.freq){
            this.setSoftwarePWMPin(pinId, message.freq, message.dc)
        }
        else {
            this.verboseLog(`Setting pin: ${pinId} to state: ${state}`)
            this.outPinsOnOff[pinId].digitalWrite(state)
            this.publish(`$state/${pinId}`, {"state": state}, {retain: true})
            //this.debugLog("Done changing pin state")
        }
    }

    setSoftwarePWMPin = (pinId, freq, dutyCycle) => {
        this.outPinsSoftwarePWM[pinId] = new Pigpio(pinId, {mode: Pigpio.OUTPUT})
        this.outPinsSoftwarePWM[pinId].pwmRange(100)
        this.outPinsSoftwarePWM[pinId].pwmFrequency(freq)
        this.outPinsSoftwarePWM[pinId].pwmWrite(dutyCycle)
        //this.debugLog("Done setting software PWM pin")
    }

    initializePin = (pinId, mode, message) => {
        //this.verboseLog("Initializing pin")
        this.managedPins[pinId] = mode
        if (mode === 1){
            this.inPins.push(new Pigpio(pinId, {mode: Pigpio.INPUT}))
        }
        else if (mode === 2){
            this.outPinsOnOff[pinId] = new Pigpio(pinId, {mode: Pigpio.OUTPUT})
        }
        else if (mode === 3){
            this.setSoftwarePWMPin(pinId, message.data.frequency, message.data.dutyCycle)
        }
    }

    stopManagingPin = (pinId) => {
        const mode = this.managedPins[pinId]
        if (mode){
            this.managedPins[pinId] = 0
            if (mode === 1){
                this.inPins.splice(this.inPins.indexOf(pinId), 1)
            }
            else if (mode === 2){
                this.outPinsOnOff[pinId] = false
            }
            else if (mode === 3){
                this.outPinsSoftwarePWM[pinId] = false
            }
            //this.debugLog(`Stopped managing pin: ${pinId}`)
        }
        
    }

    readPins = async () => {
        while (true){
            await this.sleep(1000)
            for (let i = 0; i < this.inPins.length; ++i){
                const pin = this.inPins[i]
                const value = pin.digitalRead()
                this.publishPinState(value, pin.gpio)
                // this.verboseLog(`Pin: ${pin._gpio} has value ${value}`)
            }
        }
        
    }

    readI2c = () => {
        return spawnSync("raspi-config", ["nonint", "get_i2c"], {uid: 0, gid: 0})
    }

    setI2c = (message) => {
        // Note the status here is in bash terms --> 0 means on or true
        // 1 means off or false
        const goalStatus = + !message["i2cActive"]
        const status = this.readI2c()["status"]
        if (goalStatus !== status){
            spawnSync("raspi-config", ["nonint", "do_i2c", goalStatus], {uid: 0, gid: 0})
        }
    }

    publishPinState = (state, pinId) => {
        this.publish(`$state/${pinId}`, {"state": state}, {retain:true})
    }

    sleep = (millis) => {
        // console.log(`Sleeping for ${millis} ms`)
        return new Promise(resolve => setTimeout(resolve, millis));
    }
}

gpio = new GpioApp */