const { spawn, execSync } = require('child_process')
const fs = require('fs')
const PccApp = require("../../../picockpit-library/picockpit-base-app")
const moduleBasePath = '/etc/picockpit-client/apps/com.picockpit/picontrol/modules' 

class Picontrol extends PccApp{
    constructor() {
        super("picontrol")
        this.logger.info("Constructing picontrol")
        
        this.commandsState = {}

        this.loadModules()

        this.client.on('message', (topic) => {
            this.logger.debug(`Received message, topic: ${topic}`)
            this.onControlMessage(topic)
        })
    }

    // --- Getters ---

    getRunning = (commandId) => {
        if (this.commandsState[commandId]) {
            return this.commandsState[commandId]['run']
        }
        return false
        
    }

    loadModules = async () => {
        // Loads modules non recursively, adds them as property of this.modules (module handle is name)

        // Previously the modules were sent before a connection was established with the frontend so the messages were lost
        // Now there is a 1s timeout before the modules are sent. This allows a connection to be established first.
        //This probably isn't the best solution but it works for now.
        // TODO: Make this sane by waiting for the connection instead of waiting an abitrary time.

        this.logger.verbose("Loading modules")

        this.modules = {}
        const frontendRepresentation = {}
        const modules = fs.readdirSync(moduleBasePath)
        for (const moduleName of modules) {
            const module = require(moduleBasePath + '/' + moduleName)
            // this.verboseLog(module)
            const moduleHandle = module.handle
            
            frontendRepresentation[moduleHandle] = module
            
            this.modules[moduleHandle] = {}
            for (const [commandHandle, command] of Object.entries(module['commands'])) {
                this.modules[moduleHandle][commandHandle] = command
            }
        }

        this.validateModules()

        await this.sleep(1000)

        this.logger.verbose("Publishing $meta")
        this.publish('$meta', frontendRepresentation, {retain: true})   
    }

    validateModules = () => {
        // Maybe not complete. Further validation might be added if that is required.

        let validModuleNames = []
        const forbiddenCharacters = ['#', '+', '/', '$']
        const moduleNames = Object.keys(this.modules)
        // Iterates through each module name
        for (const moduleName of moduleNames){
            // Checks if module with duplicate handle exists
            if (validModuleNames.indexOf(moduleName) > -1){
                throw new Error('Duplicate module handle')
            }
            // Checks if module handle contains forbidden characters
            for (const character of forbiddenCharacters){
                if (moduleName.includes(character)){
                    throw new Error('Forbidden character in module handle')
                }
            }
            if (typeof this.modules[moduleName] !== 'object') {
                throw new Error('No commands are included in the module')
            }
            validModuleNames.push(moduleName)
        }
        return true
    }

    onControlMessage = async (topic) => {
        // When a command is supposed to be executed the frontend publishes to:
        // pi/<pi-serial>/com.picockpit/picontrol/control/<module_name>/<command_handle>
        // This listens for those messages, extracts the module_name and the command handle,
        // checks if they exist and starts execution of the commands

        // Note: Errors coudl occur here if the command is executed before modules
        // are loaded. I suppose this could be handled by adding a sleep in here
        // or by making it wait until the modules are loaded

        const {commandHandle, moduleName} = this.getCommand(topic)
        this.logger.debug(`Command: ${commandHandle}, ${moduleName}`)
        const commandData = JSON.parse(JSON.stringify(this.modules[moduleName][commandHandle]))
        const commandId = `${moduleName}/${commandHandle}`

        if (this.isValidCommand(commandData)) {

            if (this.getRunning(commandId)) {
                this.terminateCommand(commandId)
                return
            }

            this.logger.debug(`Executing command: ${commandId}`)

            this.commandsState[commandId] = {}
            this.commandsState[commandId]['run'] = true
            this.commandsState[commandId]['stage'] = 0

            // --- Publish, that command has started running ---
            this.publishState(commandId)
            const commandArray = this.prepareCommand(commandData)

            this.runCommand(commandArray, commandId)
        }
    }

    publishState = (commandId) => {
        const run = this.commandsState[commandId]['run']
        const stage = this.commandsState[commandId]['stage']
        const finished = this.commandsState[commandId]['finished']
        const terminating = this.commandsState[commandId]['terminating']
        const killed = this.commandsState[commandId]['killed']
        const failed = this.commandsState[commandId]['failed']
        const stopped = this.commandsState[commandId]['stopped']

        const state = {run, stage, finished, terminating, killed, failed, stopped}

        this.publish(`$state/${commandId}`, state, {retain: true})
    }

    terminateCommand = (commandId) => {
        this.logger.debug(`Terminating command: ${commandId}`)
        const process = this.commandsState[commandId]['process']
        this.commandsState[commandId]['terminating'] = true
        this.publishState(commandId)
        process.kill()
        this.commandsState[commandId]['run'] = false
        this.commandsState[commandId]['terminating'] = false
        this.commandsState[commandId]['stopped'] = true
        this.commandsState[commandId]['finished'] = false
        this.publishState(commandId)
    }

    getCommand = (topic) => {
        // --- Get command handle and module handle ---

        const splitTopic = topic.split('/')
        const commandHandle = splitTopic[splitTopic.length - 1]
        const moduleName = splitTopic[splitTopic.length - 2]

        return {commandHandle, moduleName}
    }

    isValidCommand = (commandData) => {
        if (commandData) {
            return true
        }
        return false
    }

    prepareCommand = (command) => {
        if (!command['user']) {throw new Error('No user specified for command')}
        const {uid, gid} = this.getUidAndGid(command['user'])
        this.logger.debug('Uid: ', uid, 'Gid: ', gid)
        const timeout = command['timeout']

        this.logger.debug(`Command: ${command}`)

        const unformattedSubcommands = command['command']
        let commandArray = []

        if (Array.isArray(unformattedSubcommands[0])){
            this.logger.debug("Complex command")
            for (const subcommand of unformattedSubcommands){
                const formatted = this.formatCommand(subcommand, uid, timeout)
                commandArray.push(formatted)
            }
        }
        else{
            this.logger.debug("Simple command")
            const formatted = this.formatCommand(unformattedSubcommands, uid, timeout)
            commandArray.push(formatted)
        }

        return commandArray
    }

    formatCommand = (unformattedCommand, uid, timeout) => {
        this.logger.debug(unformattedCommand)
        const formattedCommand = {
            command: unformattedCommand[0],
            args: unformattedCommand.slice(1),
            uid,
            timeout
        }
        this.logger.debug(formattedCommand)
        return formattedCommand
    }

    getUidAndGid = (username) => {
        this.logger.debug('Getting uid')
        return {uid: parseInt(execSync(`id -u ${username}`, {encoding: 'utf-8'})), gid: parseInt(execSync(`id -g ${username}`, {encoding: 'utf-8'}))}

    }

    runCommand = (commandArray, commandId) => {
        let stage = this.commandsState[commandId]['stage']
        // Get information relevant to current stage from command array
        const {command, args, uid, gid, timeout} = commandArray[stage] 
        this.logger.debug("Command array before execution: ", commandArray[stage])

        // Creates process to execute command in

        const process = this.executeAsProcess(command, args, uid, gid, timeout, commandId)

        // Check if process could be created properly, or if any errors occured

        if (typeof process === "string") {
            this.commandsState[commandId]['data']['error'] = `Process could not be created successfully because of: ${process}`
            this.commandsState[commandId]['failed'] = true
            this.commandsState[commandId]['run'] = false
            this.commandsState[commandId]['data']['returncode'] = 1
            const data = {data: this.commandsState[commandId]['data']}
            this.logger.debug(this.commandsState[commandId])
            this.publish(`$state/${commandId}/0`, data, {retain: false})
            this.publishState(commandId)
            this.commandsState[commandId] = {}
            return
        }

        // If no errors occured store the process in commandState (Or create a pointer, I think that's what it is)

        this.commandsState[commandId]['process'] = process 

        // The listener waits for the command to finish running
        process.addListener('close', () => {
            // Publish results of subcommand

            const data = {data: this.commandsState[commandId]['data']}
            this.publish(`$state/${commandId}/${stage}`, data, {retain: false})
            this.logger.debug(data)

            // Check if error


            if (this.commandsState[commandId]["run"] != true) {
                const { error, signal } = this.commandsState[commandId]['data'] 
                if (error) {
                    this.commandsState[commandId]['failed'] = true
                } 
                if (signal) {
                    if (signal === "SIGKILL") {
                        this.commandsState[commandId]['killed'] = true
                    }
                    if (signal === "SIGTERM") {
                        this.commandsState[commandId]['stopped'] = true
                    }
                    else {
                        this.commandsState[commandId]['failed'] = true
                    }
                }
            }

            // Check if failed or stopped

            const { failed, stopped, killed } = this.commandsState[commandId]

            if (failed || stopped || killed) {
                this.commandsState[commandId]['run'] = false
                this.publishState(commandId)
                this.commandsState[commandId] = {}
            }
            else {
                // Check if all commands have been executed, if not carry on recursively

                this.commandsState[commandId]['stage'] = (stage + 1)
                stage = stage + 1
                if (stage < commandArray.length) {
                    this.publishState(commandId)
                    this.runCommand(commandArray, commandId)
                }
                else {
                    // --- Publish, that command has finished executing ---
                    this.commandsState[commandId]['finished'] = true
                    this.commandsState[commandId]['run'] = false
                    this.publishState(commandId)
                    this.commandsState[commandId] = {}
                }
            }
            
        })
    }

    executeAsProcess = (command, args, uid, gid, timeout, commandId) => {
        this.logger.debug("Command: ", command, "Args: ", args)
        this.commandsState[commandId]['data'] = {}
        this.commandsState[commandId]['data']['output'] = ""
        this.commandsState[commandId]['data']['error'] = ""

        try{
            const process = spawn(command, args, {timeout, killSignal: "SIGKILL", uid, gid});
            
            process.stdout.on("data", (data) => {
                this.commandsState[commandId]['data']['output'] += data
            })
            
            process.stderr.on("data", (data) => {
                this.commandsState[commandId]['data']['error'] += data
            })

            process.on('exit', (code, signal) => {
                this.commandsState[commandId]['data']['returncode'] = code
                this.commandsState[commandId]['data']['signal'] = signal

                this.logger.debug(`Command exited, code: ${code}`)
            })

            process.on('error', () => {
                this.logger.warn(`Error when executing command: ${command}, ${args}, ${uid}, ${gid}, ${timeout}, ${commandId}`)
            })

            return process
        }
        catch (e) {
            // Return error type
            return e.code
        }
    }

    sleep = async (ms) => {
        return new Promise((resolve) => {
            setTimeout(resolve, ms);
        });
    }

}

picontrol = new Picontrol()
