const { spawn } = require('child_process')
const fs = require('fs')

const PccApp = require("../../../picockpit-library/picockpit-base-app")

const moduleBasePath = '/etc/picockpit-client/apps/com.picockpit/pidoctor/modules' 

class Pidoctor extends PccApp{
    constructor() {
        super("pidoctor")

        this.logger.info('Constructing pidoctor')
        this.loadModules()

        this.runningCommands = {}

        this.client.on('message', (topic) => {
            this.logger.debug(`Recieved message, topic: ${topic}`)
            this.onControlMessage(topic)
        })
    }

    loadModules = async () => {
        // Loads modules non recursively, adds them as property of this.modules (module handle is name)

        this.logger.verbose("Loading modules")

        // Previously the modules were sent before a connection was established with the frontend so the messages were lost
        // Now there is a 1s timeout before the modules are sent. This allows a connection to be established first.
        //This probably isn't the best solution but it works for now.
        // TODO: Make this sane by waiting for the connection instead of waiting an abitrary time.

        this.modules = {}
        const frontendRepresentation = {}
        const modules = fs.readdirSync(moduleBasePath)
        for (const moduleName of modules) {
            if (moduleName == "video.json") {
                continue
            }
            const module = require(moduleBasePath + '/' + moduleName)
            const moduleHandle = module.handle
            
            frontendRepresentation[moduleHandle] = module
            
            this.modules[moduleHandle] = {}
            for (const [commandHandle, command] of Object.entries(module['commands'])) {
                if (commandHandle == "vcdbg_log_msg") {
                    this.modules[moduleHandle][commandHandle] = "test"
                } else {
                    this.modules[moduleHandle][commandHandle] = command
                }
            }
        }

        this.validateModules()

        await this.sleep(1000)

        this.logger.verbose("Publishing $meta")
        this.publish('$meta', frontendRepresentation, {retain: true})
        
    }

    validateModules = () => {
        // Maybe not complete. Further validation might be added if that is required.

        let validModuleNames = []
        const forbiddenCharacters = ['#', '+', '/', '$']
        const moduleNames = Object.keys(this.modules)
        // Iterates through each module name
        for (const moduleName of moduleNames){
            // Checks if module with duplicate handle exists
            if (validModuleNames.indexOf(moduleName) > -1){
                throw new Error('Duplicate module handle')
            }
            // Checks if module handle contains forbidden characters
            for (const character of forbiddenCharacters){
                if (moduleName.includes(character)){
                    throw new Error('Forbidden character in module handle')
                }
            }
            if (typeof this.modules[moduleName] !== 'object') {
                throw new Error('No commands are included in the module')
            }
            validModuleNames.push(moduleName)
        }
        return true
    }

    isValidCommand = (commandData) => {
        if (commandData) {
            return true
        }
        return false
    }

    getCommands = (topic) => {
        // Find out if a whole module is being called, or if only a single command
        // should be executed

        const splitTopic = topic.split('/')
        const lastPart = splitTopic[splitTopic.length - 1]

        if (this.modules[lastPart]) {
            let commandArray = []
            for (const [key, value] of Object.entries(this.modules[lastPart])) {
                commandArray.push({commandHandle: key, moduleName: lastPart})
                // this.verboseLog(`Command: ${key}, module: ${lastPart}`)
            }
            return commandArray
        }
        else {
            const moduleName = splitTopic[splitTopic.length - 2]
            // this.verboseLog(`Command: ${lastPart}, module: ${moduleName}`)
            return [{commandHandle: lastPart, moduleName}]
        }
    }

    onControlMessage = (topic) => {
        const commandArray = this.getCommands(topic)

        // Execute every single command in the array of commands
        for (const item of commandArray) {
            const {moduleName, commandHandle} = item
            const commandData = JSON.parse(JSON.stringify(this.modules[moduleName][commandHandle]))
            const commandId = `${moduleName}/${commandHandle}`

            if (this.isValidCommand(commandData)) {
                // 0. Add to runningCommands
                this.runningCommands[commandId] = {}
                // 1. Prepare command
                const preparedCommand = this.prepareCommand(commandData, commandId)
                // 2. Run command
                // TODO: Add error handling
                // 3. Format output
                this.runCommand(preparedCommand, commandHandle, commandId, commandData, moduleName)
                // 4. Remove from running commands
            }
        }
        
    }

    prepareCommand = (commandData, commandId) => {
        // Type is the return type of the command
        let { command, literal = false, timeout = 0, regex} = commandData 
        let args = null
        const commandType = typeof command

        if (commandType === "object") {
            args = command.slice(1)
            command = command[0]
        }
        else {
            if (literal) {
                args = [command]
                command = "echo"
            }
            else {
                args = [command]
                command = "cat"
            }            
        }

        return {command, args, timeout, commandId}
    }

    runCommand = (preparedCommand, commandHandle, commandId, commandData, moduleName) => {
        const process = this.executeAsProcess(preparedCommand, commandId)

        process.on('close', () => {
            this.logger.debug('Process closed')
            // The message is formatted this way so the frontend understands it
            const msg = {}
            msg['data'] = {}
            msg['data'][commandHandle] = this.runningCommands[commandId]['data']  

            this.formatOutput(commandData, commandId)

            this.publish(`$state/${moduleName}`, msg, {retain: false})
        })
    }

    formatOutput = (commandData, commandId) => {
        const { skiplines, regex } = commandData
        const output = this.runningCommands[commandId]['data']['output']
        let formattedOutput

        if (skiplines > 0){
            let temp = output.split('\n')
            temp = temp.slice(skiplines)
            formattedOutput = temp.join('\n')
        }

/*         if (regex) {
            this.verboseLog('Applying regex')
            const { expression, flags } = regex
            const exp = new RegExp(expression, flags.toLowerCase())
            this.verboseLog(formattedOutput.match(exp))
        }
 */
        return formattedOutput
    }

    executeAsProcess = (preparedCommand, commandId) => {
        const { command, args, timeout } = preparedCommand
        this.runningCommands[commandId]['data'] = {}
        this.runningCommands[commandId]['data']['output'] = ""
        this.runningCommands[commandId]['data']['error'] = ""

        this.logger.debug(`Executing: ${command}, ${args}`)

        try {
            const process = spawn(command, args, {timeout})

            //this.verboseLog('Process spawned')

            process.stdout.on("data", (data) => {
                this.runningCommands[commandId]['data']['output'] += data
            })
            process.stderr.on("error", (data) => {
                this.runningCommands[commandId]['data']['error'] += data
            })
            process.on('exit', (code, signal) => {
                this.runningCommands[commandId]['data']['returncode'] = code
                this.runningCommands[commandId]['data']['signal'] = signal

                this.logger.debug(`Command: ${command} ${args} exited. Exit code: ${code}`)

                // Temp solution for "Operation not permitted(-1) error":
                if (code === 255){
                    this.runningCommands[commandId]['data']['error'] = "Command failed, missing execution rights."
                }

            })

            return process
        }
        catch (e) {
            return e
        }
        
    }

    sleep = async (ms) => {
        return new Promise((resolve) => {
            setTimeout(resolve, ms);
        });
    }
}

pidoctor = new Pidoctor()
