const path = require('path');
const exec = require("./exec");
const files = require("./files");

/**
 * Retrieves a form by name.
 * Resolves all associated services
 *
 * @param {string} basename - The root of the service.
 * @param {string} cb - The transport "send" callback.
 */
class schemaService {
  constructor(basename, cb) {
    this.basename = basename || "conf";
    this.cb = cb;

    this.rules = {
      // isBoolean:(a) => typeof a == 'boolean', // Deprecated
      isBoolean:(a) => (a == '0' || a == '1'),  // Switch name to isNumBoolean
      isOpt:(a, cfg) => cfg.opt.includes(a),  // is it an option
      inDomain:(a, cfg) => this.inDomain(+a, cfg.domain),
    }

    this.modifiers = { // always return a string
      // toInvInt:(a) => ['1','0'][Number(a)], // shell boolean where 0 == true
      toInvInt:(a) => a, // Deprecated
      escapeChars:(a) => a.replace(/'/g, "'\\''"),
    }
  }

  /**
   * Checks if a number is within the specified domain.
   *
   * @param {number} num - The number to be checked.
   * @param {string} test - The domain string in the format "1,2,4-8".
   * @returns {boolean} `true` if the number is within the domain, otherwise `false`.
   */
  inDomain(num, test) {
    const ranges = test.split(',');

    for (const range of ranges) {
      if (range.includes('-')) {
        const [start, end] = range.split('-').map(Number);
        if (num >= start && num <= end) {
          return true;
        }
      } else {
        const value = Number(range);
        if (num === value) {
          return true;
        }
      }
    }
    
    return false;
  }

  /**
   * Wrapper function calling all schema type parsers.
   *
   * @param {string} topic - The topic.
   * @param {object} data - The request data.
   */
  serve(topic, data) {
    this.formClient(topic, data);
    this.svcClient(topic, data);
  }

  /**
   * Get forms w supplementary service data
   *
   * @param {string} topic - The topic.
   * @param {object} data - The request data.
   * @param {boolean} force - Force parsing for pushed updates. 
   */
  formClient(topic, data, force) {
    if (!force && ! (topic.includes(`${this.basename}/form/query`)))
      return;

    const formSchemas = this.getResolvedForm(data.name, data.lang);

    // Retain is needed for peculiarities of this setup
    // If the app is asleep, 1st msg will be lost.
    this.cb(`${this.basename}/form/reply`, {"name": data.name, ...formSchemas}, {retain: true}); 
    console.log("Sent form:", data.name);
  }

  /**
   * Generates a command by replacing placeholders in a template with values from an object.
   *
   * @param {string} template - The command template with placeholders.
   * @param {object} values - An object containing values to replace placeholders.
   * @returns {string} The generated command with placeholders replaced by values.
   */
  templateToCommand(template, values) {
    return template.replace(/\<(\w+)\>/g, (match, key) => values[key] || match);
  }

  /**
  * Handles a service (set) request for a specific topic.
  * Assembles default args, provided args, sanitizing inputs, modifying inputs, & executing code w optional rollbacks 
  *
  * @param {string} topic - The topic.
  * @param {object} data - Request data for the service.
  */
  svcClient(topic, data){
    if (! (topic.includes(`${this.basename}/set`)))
      return;

    // console.log("SETTING:", topic, JSON.stringify(data));
    const req = data.values;
    const services = this.getResolvedServices(Object.keys(req), true);
    // console.log("SETTING:", JSON.stringify(services));

    // for multi arg commands, get defaults if missing
    Object.keys(services).forEach(s => {
      const needArgs = Object.keys(services[s].schema);  // args the command needs
      const curArgs = {}  // Current args
      needArgs.forEach(a => {
        curArgs[a] = services[s].schema[a].val;
        if (curArgs[a] == undefined){
          console.log("Warning: No default key:", a)
        }
      }); 
      let args = {...curArgs, ...req[s]};

      for ( const a of Object.keys(req[s]) ) { // test incoming args
        const argItem = services[s].schema[a];

        // simple check
        if (String(args[a]) == '') 
          console.warn("Warning: Empty Arg", s, a);

        // rules
        const rules = argItem.rules || [];
        for ( const r of rules ) {
          if (!(r in this.rules)) {
            console.warn ("Error: Missing Rule", r);
            return;
          }
          if (!(this.rules[r](args[a], argItem.cfg))) { // if fail
            console.warn("Error: Rule Failed", s, a, r, args[a]);
            return;
          }
        }

        if ("mod" in argItem) {
          if (!(argItem.mod in this.modifiers)) {
            console.warn ("Error: Missing Modifier", argItem.mod);
            return;
          }
          args[a] = this.modifiers[argItem.mod](args[a]); // run modifier
        }
      }

      // processing for all args
      Object.keys(args).forEach(a => {
        args[a] = String(args[a]);
        // args[a] = String(args[a]).replace(/[\\]/g, match => `\\${match}`); // escape out bad chars
        args[a] = args[a].replace(/'/g, "'\\''");  // special method for single quotes
        args[a] = `'${args[a]}'`// make literal
      });

      if ('set' in services[s]) { 
        const cmd = this.templateToCommand(services[s].set, args);
        console.log ("Running:", cmd);

        // prep for "test" instead of "apply"
        // Like Ubiquiti it changes for till timeout then back
        let snapshots = [], rollbacks = [];
        if (data.type == "test") {
          snapshots = Object.keys(data.values).map(s => services[s].snapshot).filter(s => s);
          rollbacks = Object.keys(data.values).map(s => services[s].rollback).filter(s => s);
        }

        try {
          if (data.type == "test") {
            console.log("Running snapshots.")
            snapshots.forEach(c => exec.exec(c));
          }

          exec.exec(cmd);
          if (data.type == "test") {
            setTimeout(d=>{
              console.log("Running rollbacks.")
              rollbacks.forEach(c => exec.exec(c));
            }, 20000);  // test for 30sec
          }
        } catch (e) {
          console.log("Exec Errors:", JSON.stringify(e));
          this.pubStatus({timeout:-1, color:"red", text:`Exec: ${cmd} Error: ${e.stderr} Status: ${e.status}`});
        }
      } else {
        console.warn ("Warning: Missing Set Property", s, JSON.stringify(services[s]));
      }
    });
    // update new values
    console.log("SENDING UPDATED VALUES")
    if (data.update)
      this.formClient(null, data.update, true);
  }

  /**
  * Retrieves a resolved form by name with default values and language strings.
  *
  * @param {string} name - The name of the form.
  * @param {string} lang - The language identifier for fetching language strings.
  * @returns {object} The resolved form and svc schemas.
  */
  getResolvedForm(name, lang) {
    const rawForm = files.getForm(name);  

    // get needed oids
    const oids = rawForm.flatMap(d => d.children.map(e => e.svcOID));
    const uniqOids = [...new Set(oids)].map(p => path.dirname(p)); 
    const svcSchemas = this.getResolvedServices(uniqOids);

    // resolve language
    rawForm.forEach((d, i) => {
      let rfItem = rawForm[i];
      if (rfItem.i18n && rfItem.i18n.en)
        rfItem.i18n = rfItem.i18n[lang] ? rfItem.i18n[lang] : rfItem.i18n["en"];

      rfItem.children.forEach((e, j) => {
        const child = rfItem.children[j];
        const parts = child.svcOID.split("/");
        const bn = parts.slice(0, parts.length - 1).join("/");
        const dn = parts[parts.length - 1];

        // test svcOID endpoint exists
        if (!(bn in svcSchemas && dn in svcSchemas[bn])){
          console.log("No svc endpoint for:", dn, bn, child.svcOID);
          rfItem.children[j] = {};
          return;
        }

        let cfg = child.cfg;
        // Resolve to requested language
        if (cfg && cfg.title)
          cfg.title = cfg.title[lang] ? cfg.title[lang] : cfg.title["en"];

        if (cfg && cfg.props && cfg.props.label)
          cfg.props.label = cfg.props.label[lang]
            ? cfg.props.label[lang]
            : cfg.props.label["en"];

        if (child.i18n && child.i18n.en)
          child.i18n = child.i18n[lang] ? child.i18n[lang] : child.i18n["en"];

        // Add cfg.opts
        if (
          rfItem.children[j].type == "select" || rfItem.children[j].type == "combobox" ||
          rfItem.children[j].type == "radio-g"
        ) {
          const items = svcSchemas[bn][dn].cfg.opt;
          const optico = svcSchemas[bn][dn].cfg.optico;
          // TODO opt items currently lack i18n. Generally not a problem because the text is universal.
          // Ex: ["HDMI","3.5"].  CounterEx: Serial Enable Svc.  Propigate lang to arg.resolve
          if (items) {
            cfg.props.items = items;
            delete svcSchemas[bn][dn].cfg.opt;
          }; 
          if (optico) {
            cfg.optico = optico;
            delete svcSchemas[bn][dn].cfg.optico;
          } 
        }
      });
    });
    return {
      svcSchema: svcSchemas,
      formSchema: rawForm,
    };
  }

  // /**
  //  * Resolves a value and optionally executes it to return a string.
  //  *
  //  * @param {*} val - The value to be resolved.
  //  * @param {boolean} [isOpt=false] - Whether the value is an option (select item) argument.
  //  * @returns {string} The resolved string, optionally split by newlines for options.
  //  */
  // argResolve (val) {
  //   // console.log("resolve:", val, isOpt)

  //   try {
  //       return exec(val).stdout;
  //       // if (rv == "0" || rv=="1") // is this a universal?
  //       //     rv = rv == "0"?1:0;
        
  //   } catch (error) {
  //       if (typeof error == "string")
  //           console.log('Error:', error);
  //       else {
  //           console.log('Stdout:', error.stdin);
  //           console.log('Stderr:', error.stderr);
  //           console.log('Status:', error.status);
  //       }
  //       throw 'Failed exit code';
  //   }
  // }

  /**
   * Retrieves resolved services based on the requested list and resolves defaults.
   *
   * @param {string[]} svcList - The list of requested services.
   * @param {boolean} forSetting - Whether services are requested for a the setting/updating or for form resolution.
   * Setting/updating requires more information to be included.
   * @returns {object} Resolved services object.
   */
  getResolvedServices(svcList, forSetting) {
    const allSvcs = files.getServices();
    let svc = {};
    svcList.forEach(s => svc[s] = allSvcs[s]);

    let svcSchema = {};
    Object.keys(svc).forEach(d =>{
      try{
        if (svc[d].show){
          try {
            exec.exec(svc[d].show);
          } catch {
            throw "Show Test Failed"
          }
        }
        let schema = {};  
        let parent = {};

        Object.keys(svc[d].schema).forEach(e =>{
          let item = svc[d].schema[e];
          let {type, cfg} = item;

          if (forSetting) { // setting (running a command) requires more items
            schema[e] = {type, cfg};
            if ("rules" in item) schema[e].rules = item.rules;
            if ("mod" in item) schema[e].mod = item.mod;
            if ("set" in svc[d]) parent.set = svc[d].set 
            if ("snapshot" in svc[d]) parent.snapshot = svc[d].snapshot 
            if ("rollback" in svc[d]) parent.rollback = svc[d].rollback 
          } else
            schema[e] = {type, cfg};

          let s = schema[e];
          
          // If cfg.opt needs to be resolved
          if ('cfg' in s && s.cfg && 'opt' in s.cfg) {
            if ('opt' in s.cfg)
              s.cfg.opt = exec.execLog(s.cfg.opt).split("\n");
            if ('optico' in s.cfg)
              s.cfg.optico.forEach(o => {
                o.list = exec.execLog(o.list).split("\n");
              });       
          }
          // poll current value
          if ('get' in item) s.val = exec.execLog(item.get);
    
          if (!forSetting) {  // modify them for the frontend, for setting leave them be
            // if(s.type == "boolean") s.val = !!+s.val;
            if(s.type == "radio-g") s.val = +s.val;
          }
        });
        if (forSetting) {
          svcSchema[d] = parent;
          svcSchema[d].schema = schema;
        } else
          svcSchema[d] = schema;

      } catch (err) {
        console.log("Omitting svc", d, err);
      }
    });
    
    return svcSchema;
  }
  
  /**
   * PubStatus
   * @description Publish status object with color, text, and timeout to the frontend (currently a vuetify snackbar)
   * @param {Object} info 
   */
  pubStatus(info) {
    this.cb(`${this.basename}/info`, info); 
  }
}

module.exports = schemaService;
