// Used for communicating with the Picockpit API

const request = require("request")
const { Buffer } = require('buffer')
const crypto = require('node:crypto')
const util = require("node:util")

const asyncRequest = util.promisify(request)

module.exports = class API{
    constructor(clientCredentials, apiLocation){
        // console.log("Api being constructed")
        this.clientCredentials = clientCredentials,
        this.apiLocation = apiLocation

        // console.log(clientCredentials)
        // console.log(apiLocation)
    }

    apicall = (suburl="/", method="GET", data="") => {
        //console.log("Calling API")
        
        let success = false
        let triesLeft = 5
        let requestObject

        while (!success && triesLeft > 0) {
            if (triesLeft != 5) {
                console.log(`Making API call, try: ${6-triesLeft}`)
            }
            triesLeft -= 1

            const fullurl = this.apiLocation + suburl
            
            const headers = this.getAuthenticationHeaders()
            headers["Accept"] = "application/json"
            headers["Content-type"] = "application/json"
    
            const netloc = this.apiLocation.split("/")[2]
            const pathPart1 = this.apiLocation.split("/")[3]

            const hco_hmac = this.getHmac(
                headers["nonce_ts"],
                method, 
                netloc, // Netloc
                "/" + pathPart1 + suburl, // path
                "", // query
                data, // body
            )
    
            headers["hmac"] = hco_hmac
    
            //console.log("Starting request")
            //console.log(headers)
    
            /* console.log(util.inspect(
                {
                    method,
                    url: fullurl,
                    json: data,
                    headers
                }, false, null
            
                )
            ) */
            requestObject = asyncRequest(
                {
                    method,
                    url: fullurl,
                    json: data,
                    headers
                }
            )
    
            if (!requestObject.error){
                success = true
            }
        }

        //console.log("Request completed (still awaiting answer, returned promise)")
        
        return requestObject
    }

    getAuthenticationHeaders = () => {
        const { prefix, serial, hash_c_two } = this.clientCredentials
        const headers = {
            "nonce_ts": Math.floor(Date.now() / 1000).toString(),
            prefix,
            "bcm_serial": serial,
            hash_c_two,
        }

        return headers
    } 

    getHmac = (nonce_ts, accessMethod, netloc, path, query, body) => {
        const hcoBytes = Buffer.from(this.clientCredentials["hash_c_one"], "hex") 
        const canurl = netloc + path + "?" + query

        if (!body){
            body = '""'
        }
        else if (body !== ""){
            body = JSON.stringify(body).replaceAll('":', '": ').replaceAll(",", ", ")//.replaceAll("\\", "\\\\")
        }

        const message = nonce_ts + "!" + accessMethod + "!" + canurl.toLowerCase() + "!" + body 

        const hcoHmacEncoder =  crypto.createHmac("sha256", hcoBytes)
        hcoHmacEncoder.update(message)
        const hco_hmac = hcoHmacEncoder.digest('hex')

/*         console.log("----------------------")
        console.log(JSON.stringify(hcoBytes))
        console.log(message)
        console.log("----------------------")
 */
        return hco_hmac
    }

    getMqttCredentials = async () => {
        
        //console.log("------------------------- API Call -------------------------")
        console.log("Getting the MQTT credentials")

        //console.log(this.clientCredentials)

        const requestObject = await this.apicall(
            "/pi/" + (this.clientCredentials.serial).toString() + "/mqtt/credentials",
            "GET",
        )

        const apiAnswer = JSON.parse(requestObject.body)

        //console.log(`Apianswer: ${apiAnswer}`)

        if (apiAnswer.error){
            console.log('\x1b[31m%s\x1b[0m', "Error: Creating the MQTT credentials didn't work.")
            console.log('\x1b[31m%s\x1b[0m', apiAnswer)
            console.log('\x1b[31m%s\x1b[0m', "       1. Check if the Pi is already registered to your account - delete it in the web interface if this is the case.")
            console.log('\x1b[31m%s\x1b[0m', "       2. Restart this installation process.")
            console.log('\x1b[31m%s\x1b[0m', "       3. Please contact our support: support@pi3g.com")
            process.exit(1)

            throw new Error("Error when making API-call (MQTT credentials)")
        }

        else if (apiAnswer.password.length === 50){
            return apiAnswer
        }
    
    }

    addPi = async (pidata) => {

        //console.log("------------------------- API Call -------------------------")
        console.log("Adding this Raspberry Pi to the database")
        //console.log(pidata)

        const requestObject = await this.apicall(
            "/pi",
            "POST",
            pidata
        )

        const apiAnswer = requestObject.body

        // console.log(`Apianswer: ${util.inspect(apiAnswer, false, null)}`)
        if (apiAnswer.error){
            console.log('\x1b[31m%s\x1b[0m', "Error: Couldn't add the Pi to database")
            console.log('\x1b[31m%s\x1b[0m', "       1. Check if the Pi is already registered to your account - delete it in the web interface if this is the case.")
            console.log('\x1b[31m%s\x1b[0m', "       2. Restart this installation process.")
            console.log('\x1b[31m%s\x1b[0m', "       3. Please contact our support: support@pi3g.com")
            process.exit(1)
    
            //throw new Error("Error when making API-call (add pi) - Make sure the Pi isn't already registered on your account!")
        }

        return apiAnswer

    }
}