// Everything in this file is used by both the setup mode and the daemon mode

// TODO: Make more expressive variable names
// I have no clue what hash_c_one or hSalt mean

// TODO: Change name of this class

const fs = require('fs')
const crypto = require('node:crypto')
const { Buffer } = require("buffer")
const API = require('./api')
const fernet = require('fernet')
const { exit } = require('process')
const winston = require('winston');
const { combine, timestamp, printf, colorize, align, json } = winston.format;
const path = require('path');


module.exports = class UniversalFunctions{
    constructor(args, daemonMode){
        
        this.CONFIG_FILE = '/etc/picockpit-client/picockpit-client.config.json'
        this.DEFAULT_CONFIG_FILE = './config/default_conf_picockpit.json'
        
        this.args = args
        this.daemonMode = daemonMode

        // Class variables
        this.config
        this.serial
        this.encodedSerial
        
        this.cred_salt
        this.decryptedClientCredentials

        this.api

        this.mqttSelfsigned

        this.printLogLevel
        this.saveLogLevel
        
        this.readConfig()

        this.logger = winston.createLogger({
            transports: [
                new winston.transports.Console({
                    levels: { silent: 0, error: 1, warn: 2, info: 3, http: 4, verbose: 5, debug: 6, silly: 7 },
                    level: this.printLogLevel || "info",
                    format: combine(
                        colorize({all: true}),
                        timestamp({
                            format: 'hh:mm:ss',
                        }),
                        align(),
                        printf((info) => `[${info.timestamp}] ${info.level}: ${info.message}`)
                    ) 
                }),
                new winston.transports.File({
                    levels: { silent: 0, error: 1, warn: 2, info: 3, http: 4, verbose: 5, debug: 6, silly: 7 },
                    level: this.saveLogLevel || "silent",
                    filename: `/var/log/picockpit-client/core.log`,
                    format: combine(
                        timestamp({
                          format: 'YYYY-MM-DD hh:mm:ss.SSS',
                        }),
                        json()
                    ),
                })
            ],
        });
        
        // Initialisation functions

        
        this.getSerial()

        if (daemonMode){
            this.setUpCredentials()
            this.setUpAPIConnection()
        }
    }

// --- Initialisation

    readConfig = () => {
        // Read the config json file
        
        // this.logger.verbose("Reading config")

        if (fs.existsSync(this.CONFIG_FILE)){
            // this.logger.verbose("Config found")
            const rawConfig = fs.readFileSync(this.CONFIG_FILE)
            this.config = JSON.parse(rawConfig)
        }
        else if (fs.existsSync(this.DEFAULT_CONFIG_FILE)){
            // this.logger.verbose("No config found, using default config")
            const rawConfig = fs.readFileSync(this.DEFAULT_CONFIG_FILE)
            this.config = JSON.parse(rawConfig)
        }
        else {
            // this.logger.error("No default config file found")
            throw {"ERROR": 'ERROR: No default config file found'}
        }

        // Check if config contains bcmhash_salt
        if (this.config['bcmhash_salt'] === '?'){
            // Get bcmhash salt from secrets safe?
            // Write a settings file
        }

        this.printLogLevel = this.config["printLogLevel"]
        this.saveLogLevel = this.config["saveLogLevel"]

        this.mqttSelfsigned = true
        this.mqttSelfsigned = this.config.mqtt_selfsigned
        if (this.args.selfsigned){
            this.mqttSelfsigned = this.args.mqttSelfsigned
        }
    }

    getSerial = () => {
        // TODO: Add option of user input serial override
        this.logger.verbose("Getting serial")

        if (this.args.serial){
            this.serial = this.args.serial
        }
        else{
            let raspiSerial = fs.readFileSync('/sys/firmware/devicetree/base/serial-number', 'utf-8', (err) => {
                if (err) {
                    this.logger.error("Failed to read raspi serial number")
                    throw {"ERROR": 'ERROR: Reading raspi serial number failed'}
                }
            })
            raspiSerial = JSON.stringify(raspiSerial)
            raspiSerial = raspiSerial.slice(1, -7)
            
            this.serial = raspiSerial
            this.logger.verbose("Serial is: ", this.serial)
        }
        
        this.setEncodedSerial()
    }

    setUpCredentials = () => {
        this.logger.verbose("Reading + setting up credentials")
        
        if (this.config["apikeys"][this.encodedSerial]){
            
            const relevantConfig = this.config["apikeys"][this.encodedSerial]
            this.cred_salt = relevantConfig["cred_salt"]

            const prefix = this.symmetricDecode(this.cred_salt, relevantConfig["prefix"])
            const hash_c_one = this.symmetricDecode(this.cred_salt, relevantConfig["hash_c_one"])
            const hash_c_two = this.symmetricDecode(this.cred_salt, relevantConfig["hash_c_two"])

            this.decryptedClientCredentials = {
                "serial": this.serial,
                prefix,
                hash_c_one,
                hash_c_two
            }

            this.logger.verbose("Finished setting up credentials")
        }
        else{
            this.logger.error("ERROR: No credentials exist for this pi")
            exit()
        }

    }

    setUpAPIConnection = () => {
        this.logger.verbose("Setting up API connection")
        //console.log("DecryptedClientCredentials")
        //console.log(this.decryptedClientCredentials)

        if (this.decryptedClientCredentials){
            this.api = new API(this.decryptedClientCredentials, this.config["api_location"])
            this.logger.verbose("Finished setting up API connection")
        }
        else{
            this.logger.error("ERROR: Failed to set up API connection because of missing client credentials")
            exit()
        }
    }

// --- crypto ---

    setEncodedSerial = () => {
        this.logger.verbose("Encoding serial")
        
        try {
            const temp = crypto.pbkdf2Sync(
                this.serial,
                this.config.bcmhash_salt,
                100000,
                64,
                'sha512'
            )
            this.encodedSerial = temp.toString('base64')
            this.logger.verbose("Encoded serial is:", this.encodedSerial)
        }
        catch (err) {
            console.error(err)
            this.logger.error(`ERROR: Failed to encode serial, error: ${err}`)
            exit()
        }
    }

    generateEncryptionKey = (hSalt) => {
        try{
            this.logger.verbose("Generating encryption key")
            const password = this.serial;
            const salt = Buffer.from(hSalt, 'hex');
            const derivedKey = crypto.pbkdf2Sync(
                password,
                salt,
                100000,
                32,
                'sha256'
            );
            const key = derivedKey.toString('base64').replace('+','-').replace('/','_');
    
            this.logger.verbose("Encryption key: ", key)
            return key
        }
        catch (err) {
            this.logger.error(`ERROR: Error generating encryption key, error: ${err}`)
            exit()
        }
    }

    symmetricDecode = (hSalt, hToken) => {
        this.logger.verbose("Decoding")

        try{
            const key = this.generateEncryptionKey(hSalt)
    
            const fernetSecret = new fernet.Secret(key);
    
            // Input token is treated as hex input and converted to utf8
            const d_h_token = Buffer.from(hToken, 'hex').toString('utf-8');
    
            const token = new fernet.Token({
                secret: fernetSecret, 
                token: d_h_token,
                ttl: 0
            });
            const decodedToken = token.decode();
    
            this.logger.verbose("Finished decoding")
            return decodedToken;
        }
        catch (err) {
            console.error(err)
            this.logger.error(`ERROR: Error decoding, error: ${err}`)
            exit()
        }

    }

    symmetricEncode = (hSalt, payload) => {
        this.logger.verbose("Encoding")

        try {
            // TODO: Add salt generation if no salt is entered
            const key = this.generateEncryptionKey(hSalt)
    
            const secret = new fernet.Secret(key)
            // Not quite sure why date and the array of integers is used, but it works
            // so I won't touch it
            const token = new fernet.Token({
                secret: secret,
                time: Date.parse(1),
                iv: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
            })
            const encoded = token.encode(payload)
    
            const hexRet = {
                token: Buffer.from(encoded, 'utf-8').toString('hex'),
                salt: hSalt
            }
    
            this.logger.verbose("Finished encoding")
    
            return hexRet 
        }
        catch(err){
            console.error(err)
            this.logger.verbose(`ERROR: Error encoding, error: ${err}`)
            exit()
        }

    }

// --- other ---

    updateSettings = () => {
        this.writeSettings()
        this.setUpCredentials()
    }

    writeSettings = () => {
        // Write config to JSON file
        // TODO: Error if CONFIG_FILE path does not exist --> Handle this later

        try{
            const path = this.CONFIG_FILE
            const data = JSON.stringify(this.config, null, 2)
    
            fs.writeFileSync(path, data)
        }
        catch(err){
            this.logger.verbose(`ERROR: Error writing settings, error: ${err}`)
            exit()
        }
    }

    async sleep(millis){
		return new Promise(resolve => setTimeout(resolve, millis));
	}
}