// The functions here are required for normal usage of the Picockpit client, not the setup

const UniversalFunctions = require("./../lib/universalFunctions")
const { Worker } = require("node:worker_threads")
const mqtt = require("mqtt")
const msgpack = require("@msgpack/msgpack")
const aedes = require("aedes")
const net = require("net")
const fs = require("fs")
const util = require("util")
const { exec } = require('child_process');


module.exports = class DaemonMode{
    constructor (args){
        this.uniFunctions = new UniversalFunctions(args, true)
        
        this.uniFunctions.logger.info("Starting in daemon mode")

        // Connection between Picockpit server and core

        this.mqttClient
        this.topicRoot = `pi/${this.uniFunctions.serial}`

        // Connection between apps and core

        this.aedes
        this.server
        this.app_ports = {}

        this.runningApps = {}

        const fs = require('fs');
        const path = require('path');
        
        const folderPath = '/etc/picockpit-client/svc.d'; // Change this to the path of your folder

        /*
        # Newest Version of Picockpit sets a 'hold' -> it doesn't get updated with a normal "sudo apt upgrade". 
        # This also gets mentioned on a potential upgrade in the future, so the user (should) know that it is happening:
        
            # Reading package lists... Done
            # Building dependency tree       
            # Reading state information... Done
            # Calculating upgrade... Done
            # The following packages have been kept back:
            #  picockpit-client
            # 0 upgraded, 0 newly installed, 0 to remove and 1 not upgraded.
        
        
        # 2. This allows the user to execute sudo apt upgrade scripts without having to worry about the client restarting,
           # us making special cases inside the client, etc. as discussed.
        # 3. Long term (Once we push an update) we will add a 'special update script' to the users JobQueueMgr to allow
           # them to do an update specific for picockpit.
        #=> This is the way this functionality will be handled until at some point we have a more robust logger/etc. 
        #   which would maybe allow persistent logs for self-upgrades etc. But until then this is the way of least 
        #   resistance which still gives the users the ability to use 'Pi-Upgrade-scripts' without the danger of losing relevant information etc.
        */
       
        fs.readFile('/etc/picockpit-client/picockpit-client.config.json', 'utf8', (err, data) => {
            if (err) {
                console.error('Error reading file:', err);
                return;
            }
            try {
                // Check if the field 'hold_updates' exists and its value is true
                // Parse the JSON data
                const jsonData = JSON.parse(data);
                if (jsonData.hasOwnProperty('hold_updates') && jsonData.hold_updates == false) {
                    console.log('The field "hold_updates" exists and its value is false.');
                } else {
                    console.log('The field "hold_updates" either does not exist or its value is true.');
                    exec("sudo apt-mark hold picockpit-client", (error, stdout, stderr) => {
                        if (error) {
                            console.error(`Error executing command: ${error.message}`);
                            return;
                        }
                        if (stderr) {
                            console.error(`Command STDERR: ${stderr}`);
                            return;
                        }
                        console.log(`Command output: ${stdout}`);
                    });    
                }
            } catch (error) {
                console.error('Error parsing JSON:', error);
            }
        });
        
        fs.readdir(folderPath, (err, files) => {
          if (err) {
            console.error('Error reading configuration folder:', err);
            return;
          }
        
          files.forEach(file => {
            const filePath = path.join(folderPath, file);
            fs.readFile(filePath, 'utf8', (err, data) => {
              if (err) {
                console.error(`Error reading configuration file ${file}:`, err);
                return;
              }
        
              try {
                const jsonData = JSON.parse(data);
                this.app_ports[jsonData["name"]] = jsonData["com_port"];
              } catch (parseError) {
                console.error(`Error parsing JSON from file ${file}:`, parseError);
              }
            });
          });
        });          

        this.setUpMqttConnection()
        this.setUpAedesServer()

        // Start apps like this for the moment 
        // (Look commented code in handle messages from frontend for how it
        // could be done)

        const apps = [
            "pistats", 
            "pidoctor",
            "picontrol",
            // "settings",
            // "streamvideo",
            // "co2sensor",
            // "gpio",
            // "editor",
            // "bme688airmeter"
        ]

        apps.forEach(app => this.startApp(app))
        
    }

    // Connection between Picockpit server and core

    setUpMqttConnection = () => {
        this.uniFunctions.logger.verbose("Establishing mqtt connection (between daemon and server)")
        
        const { mqtt_server, mqtt_port, mqtt_path } = this.uniFunctions.config
        const { clientId, username, password } = this.readMqttCredentials()
    
        const mqttUri = `wss://${mqtt_server}:${mqtt_port}${mqtt_path}`
        const connectionOptions = {
            clientId,
            username, 
            password,
            reconnectPeriod : 30000,
            will:{
                topic:`${this.topicRoot}/$state`,
                payload: this.prepareMessage({"onlineState": "offline"}),
                retain: true
            }
        }
        if (this.uniFunctions.mqttSelfsigned == true) {
            connectionOptions.rejectUnauthorized = false
        }

        this.mqttClient = mqtt.connect(
            mqttUri,
            connectionOptions
        )

        this.mqttClient.on("connect", () => {
            console.log('\x1b[42m%s\x1b[0m', "Connection successful")

            this.publishToFrontend(`${this.topicRoot}/$state`, {"onlineState": "online"}, {retain: true})
            // TODO: Read version automatically - set startup time - set connectionCounter.
            const msg_version = {"version": "3.0.1","startuptime": "","connectionCounter": "1", "run": "True"}
            this.publishToFrontend(`${this.topicRoot}/com.picockpit/core/$state/client`, msg_version, {retain: true})
            this.subscribeToMqttTopics()
        })

        this.mqttClient.on("error", (err) => {
            console.log("Error: ", err)
        })

        this.handleMessagesFromFrontend()
    }

    async fetchHandler(msg) {
        const nonce = msg.nonce;
        const url = msg.url;
        const payload = msg.payload;
        // console.log(url, JSON.stringify(payload));
        const mf_res = `pi/${this.uniFunctions.serial}/com.picockpit/fetch/res`;
        
        try {
          const splitUrl = url.split("/");
          console.log("1")
          const svc = splitUrl[3];
          console.log("12")

          const path = splitUrl.splice(4).join("/");
          const port = this.app_ports[svc];
          const res = await fetch("http://127.0.0.1:" + port + "/" + path, payload);
          const text = await res.text();
          console.log("13")

          this.publishToFrontend(mf_res, {nonce, res:{headers:{}, status:res.status, text}});
          console.log("14")

        } catch (e) {
            console.log(e);
            this.publishToFrontend(mf_res, {nonce, res:{headers:{}, status:500, text:`{"status":"500", "message":"Client Error: ${e}"}`}});
        }
    }

    handleMessagesFromFrontend = async () => {
        this.mqttClient.on("message", (topic, message) => {
            var type = 'internal_mqtt';
            const fetch_req = /^pi\/[a-zA-Z0-9]+\/com\.picockpit\/fetch\/req$/;
            const fetch_res = /^pi\/[a-zA-Z0-9]+\/com\.picockpit\/fetch\/res$/;
            if (fetch_req.test(topic)) {
                type = 'fetch';
            } else if (fetch_res.test(topic)) {
                // This seems to only happen in weird instances where the frontend complains about the jobqueuemgr not being reachable.
                // This is probably a place where the client should check in on the jobqueuemgr daemon and fix stuff if necessary.
                // TODO.
                type = 'ignore';
            }
            switch (type) {
                case 'fetch':
                    this.fetchHandler(msgpack.decode(message));
                    break;
                case 'internal_mqtt':
                    const appName = this.getAppFromTopic(topic)
                    // this.uniFunctions.verboseLog("App name: ", appName)
                    if (topic.includes("picontrol")) {
                        this.startApp(appName);
                    }
                    if (topic.includes("pidoctor")) {
                        this.startApp(appName);
                    }

                    if (appName === "core"){
                        // This is not good and only a temporary solution, until I
                        // understand how the gpio app is supposed to be started
                    
                        if (topic.includes("gpio")){
                            this.startApp("gpio")
                        }
                    } else if (appName === "co2sensor") {
                        // We ignore CO2 for now -> Segfault.
                    }
                    else{
                        if (topic.includes("/pistats/$state")){
                            // TODO: This is obviously super ugly - we need to fix the 'client gets its own messages' thing.
                            this.uniFunctions.logger.debug("Filter Pistats self-renew.")
                        } else {
                            this.uniFunctions.logger.debug("Renewing ", appName )
                            this.startApp(appName)    
                        }
                    }
                    
                    try{
                        if (message.length > 0){
                            message = JSON.stringify(msgpack.decode(message))
                        }
                    }
                    catch(err){
                        this.uniFunctions.logger.debug(`Failed to unpack message, topic: ${topic}, message: ${JSON.stringify(util.inspect(message, {depth: null}))}, error: ${err}`)
                        // console.error(err)
                        // console.error("Failed to unpack message")
                    }
                    this.publishToApps(topic, message)
                    break;
                case 'ignore':
                    break;
                default:
                    console.error("Failed to differentiate between Fetch and internal MQTT")
                    break;

            }
        })
    }

    readMqttCredentials = () => {
        this.uniFunctions.logger.verbose("Reading mqtt credentials")
        
        const encodedSerial = this.uniFunctions.encodedSerial
        const relevantConfig = this.uniFunctions.config["apikeys"][encodedSerial]
        
        if (relevantConfig){
            const salt = relevantConfig["mqtt_salt"]
            const encMqttPassword = relevantConfig["mqtt_password"]

            const mqttCredentials = {
                clientId: this.uniFunctions.serial, 
                username: this.uniFunctions.decryptedClientCredentials["prefix"],
                password: this.uniFunctions.symmetricDecode(salt, encMqttPassword)
            }

            this.uniFunctions.logger.verbose("Finished reading mqtt credentials")
            return mqttCredentials
        }
        else {
            this.uniFunctions.logger.error("No mqtt credentials saved in config")
            throw {"ERROR": "No mqtt credentials saved in config"}
        }
    }

    subscribeToMqttTopics = () => {
        this.mqttClient.subscribe(
            `${this.topicRoot}/com.picockpit/#`)
    }

    prepareMessage = (message) => {
        // Prepares messages before they are sent to the server.
        message.$ts = new Date().getTime() / 1000
        message.$client = `pi/${this.uniFunctions.serial}`
        
        return msgpack.encode(message)
    }

    publishToFrontend = (topic, message, options) => {
        // TODO: Maybe simplify the topic. (Automatically add topicRoot for example)

        message = this.prepareMessage(message)
        this.uniFunctions.logger.verbose("Publishing to topic: " ,topic)
        //console.log(message)
        this.mqttClient.publish(topic, message, options)
    }

    // Connection between apps and core

    setUpAedesServer = () => {
        // This server serves as the connection between the apps and the core

        this.uniFunctions.logger.verbose("Setting up aedes-server")

        try{
            this.aedes = new aedes.Server()
            this.server = net.createServer(this.aedes.handle)
        }
        catch (err) {
            console.error(err)
            this.uniFunctions.logger.error(`Failed to create Aedes server, error: ${err}`)
            throw {"ERROR": "Falied to create Aedes server"}
        }
        
        const port = 1883

        this.server.listen(port, () => {
            this.uniFunctions.logger.debug("Server started listening on port ", port) 
        })

        this.aedes.on("client", (clientId) => {
            this.uniFunctions.logger.debug(`Client connecting: ${clientId}`)
        })

        this.aedes.on("clientReady", (clientId) => {
            this.uniFunctions.logger.debug(`Client connected: ${clientId}`)
        })

        this.aedes.on("clientDisconnect", (clientId) => {
            this.uniFunctions.logger.debug(`Client disconnected: ${clientId}`)
        })

        this.aedes.authenticate = (client, username, password, callback) => {
            callback(null, username === "test")
        }

        this.handleMessagesFromApps()
    }

    publishToApps = async (topic, message) => {
        // Messages recieved via mqtt from the picockpit server are decoded in 
        // this file and then published to the app it belongs to using an other
        // mqtt connection. This mqtt connection is local. An aedes server run
        // in this file is used to publish the decoded messages.
        
        // Question about topic: Would it make sense to change the topic of all
        // messages published by the aedes server to make subscribing more
        // simple? That way it would be clear exactly which messages are
        // intended for the apps, and which are supposed to be published to the
        // frontend.

        await this.uniFunctions.sleep(1000)

        // console.log("---- Publishing to app ---- ")
        // console.log("Topic: ", topic)
        // console.log("Message: ", message)
        this.aedes.publish({topic, payload: message})
    }

    handleMessagesFromApps = () => {
        this.aedes.on("publish", (packet, client) => {

            if (packet.topic.includes("internal")){
                packet.topic = packet.topic.substring(9)
                // console.log(packet.topic)
            
                if (packet.topic.includes("$SYS")){
                    // console.log("Heartbeat or registering of new client. --> Handle soon?")
                    // console.log(packet.topic)
                    // console.log(packet.payload)
                }
                else{
                    // this.uniFunctions.verboseLog(`--- App: published something ---`)
                    // this.uniFunctions.verboseLog(`Topic: ${packet.topic}`)
                
                    const topic = `${this.topicRoot}/${packet.topic}`
                    const payload = JSON.parse(packet.payload)
                    const retain = { retain: packet.retain }

                    if (this.isObject(payload)){
                        // This is necessary because sometimes a payload isn't an object
                        // but an integer. I don't know what that integer is supposed
                        // to do thogh 
                        this.publishToFrontend(topic, payload, retain)
                    }
                    else{
                        // Not quite sure what to do in this case
                        // this.uniFunctions.verboseLog("Not an object, cant give ts")
                    }
                }
            }
            else {
                // console.log("Message wrongly attributed to app")
                // console.log(packet.topic)
                // console.log(packet.payload.toString())
            }

        })
    }

    // Other

    startApp = (appName) => {
        // Problem: The pccApp tries reading a config file

        // Make visible in "runningApps" if app is also connected
        console.log("Starting app: ", appName)

        if (!this.runningApps.hasOwnProperty(appName)){
            // this.uniFunctions.debugLog(`Starting App: ${appName}`)

            const appRoot = "./../picockpit-apps"
            const appDev = "com.picockpit"
            var fullPath = `${appRoot}/${appDev}/${appName}/${appName}`

            // this.uniFunctions.verboseLog(`Main app file: ${fullPath}`)

            this.uniFunctions.logger.verbose(`Starting App: ${fullPath}`)
            this.uniFunctions.logger.debug(process.cwd())
            if (fs.existsSync(fullPath + '.js')){
                fullPath = fullPath + '.js'
            } else if (fs.existsSync(fullPath + '.jsc')) {
                fullPath = fullPath + '.jsc'
            } else {
                this.uniFunctions.logger.debug(`App: "${appName}" does not exist`)    
            }

            const app = new Worker(fullPath)

            this.runningApps[appName] = {
                app,
                timeout: null,
            }, 

            this.setAppTimeout(appName, 60)

            this.runningApps[appName]["app"].on("error", (err) => {
                this.uniFunctions.logger.debug(`ERROR - ${appName}, error: ${err}`)
                console.log(err)
            })

            this.runningApps[appName]["app"].on("exit", (exitCode) => {
                this.uniFunctions.logger.verbose(`Closing app "${appName}" with exit code "${exitCode}"`)
                const timeoutId = this.runningApps[appName]["timeout"] 
                clearTimeout(timeoutId)
                delete this.runningApps[appName]
            })
        }
        else {
            this.uniFunctions.logger.verbose(`App is already running --> Resetting timeout (${appName})`)
            const timeoutId = this.runningApps[appName]["timeout"]
            clearTimeout(timeoutId)
            this.setAppTimeout(appName, 60)
        }
    }

    setAppTimeout = async (appName, seconds) => {
        this.runningApps[appName]["timeout"] = setTimeout(this.terminateApp, seconds * 1000, appName)
    }

    terminateApp = (appName) => {
        this.uniFunctions.logger.verbose(`Terminating ${appName} because of timeout`)

        console.log("Killing app: ", appName);
        
        // Initiate termination
        const success = this.runningApps[appName]["app"].terminate()

        if (success) {
            this.uniFunctions.logger.verbose('Worker termination initiated successfully.');
            // Listen for the 'exit' event
            this.runningApps[appName]["app"].on('exit', (code) => {
                this.uniFunctions.logger.verbose(`${appName} exited with code ${code}.`);
                // Additional cleanup or handling if needed
            });
        } else {
            this.uniFunctions.logger.error(`Failed to initiate ${appName} termination.`);
        // Handle the failure accordingly
        }
    }

    getAppFromTopic = (topic) => {
        return topic.split("/")[3]
    }

    isObject = (a) => {
        return (!!a) && (a.constructor === Object);
    }


    /*sleep = async (ms) => {
        return new Promise((resolve) => {
            setTimeout(resolve, ms);
        });
    }*/

}
