// Everything here is used when credentials are first set up, mqtt credentials are generated etc.  

const UniversalFunctions = require("./../lib/universalFunctions")
const prompt = require("prompt-sync")()
const mqtt = require("mqtt")
const crypto = require('node:crypto')
const { Buffer } = require("buffer")
const fs = require('fs')
const util = require('node:util')

const fetch = require("node-fetch")

module.exports = class SetupMode{
    constructor(args){
        this.uniFunctions = new UniversalFunctions(args, false)
        if (this.uniFunctions.mqttSelfsigned == true) {
            process.env["NODE_TLS_REJECT_UNAUTHORIZED"] = 0;
        }

        this.uniFunctions.logger.info("Starting in setup mode")

        this.args = args
        this.uniFunctions.logger.verbose(`Setup args: ${JSON.stringify(util.inspect(this.args, {depth: null}))}`)
    }

    // --- OTK ---

    useOtk = async (otk, piName) => {
        this.uniFunctions.logger.info("Using OTK for setup")

        const response = await fetch('https://picockpit.com/api/v2.0/otk/newapikey', {
            method: 'POST',
            body: JSON.stringify({
                otk,
                piName
            }),
            headers: {
                'Content-type': 'application/json; charset=UTF-8'
            }
        })
        const responseJson = await response.json()
        
        if (!responseJson.error){
            return this.newApiKey(responseJson.apikey)
        }
        else {
            this.uniFunctions.logger.error(`ERROR: Error getting Apikey using OTK, API response: ${responseJson}`)
            throw {"ERROR": "Error getting Apikey using OTK"}
        }

    }

    // --- Adding Api key ---

    newApiKey = (key) => {

        this.uniFunctions.logger.info("Adding API key")

        let keyValid = false
        let apiKey = ""
        try {

            if(key){
                apiKey = key
                keyValid = true
            }
            else if(this.args.apikey){
                apiKey = this.args.apikey
                console.log("Key", apiKey)
                keyValid = this.checkApiKeyValidity(apiKey)
    
                if(!keyValid){
                    console.log('\x1b[31m%s\x1b[0m', "Error: Invalid Key")
                    console.log('\x1b[31m%s\x1b[0m', "       1. Check for typing errors.")
                    console.log('\x1b[31m%s\x1b[0m', "       2. Create a new API-Key in the web-interface.")
                    console.log('\x1b[31m%s\x1b[0m', "       3. If those don't work please contact us at: support@pi3g.com")
                }
            }
            else{
                while(!keyValid){
                    console.log('\x1b[35m%s\x1b[0m', "Enter API key: ")
                    apiKey = prompt("       ")
                    keyValid = this.checkApiKeyValidity(apiKey)
                
                    if(!keyValid){        
                        console.log('\x1b[31m%s\x1b[0m', "Error: Invalid Key")
                        console.log('\x1b[31m%s\x1b[0m', "       1. Check for mistakes in copying or typing the key.")
                        console.log('\x1b[31m%s\x1b[0m', "       2. Create a new API-Key in the web-interface.")
                        console.log('\x1b[31m%s\x1b[0m', "       3. If those don't work please contact us at: support@pi3g.com")
                        }
                }
            }
        }
        catch(err){
            this.uniFunctions.logger.error("ERROR: Error while entering API key")
            // throw new Error("ERROR: Error while entering API key")
            throw {"ERROR": "Error while entering API key"}
        }


        if (keyValid){
            
            this.uniFunctions.logger.info("Api key is valid")        
            // Check if API entry already exists for this pi
            if (this.uniFunctions.config["apikeys"][this.uniFunctions.encodedSerial]){
                this.uniFunctions.logger.info("An API entry already exists for this pi")
                const replace = prompt("Do you want to replace it? (y/N) ")
                console.log('\x1b[0m')
                if (replace !== "y"){
                    console.log('\x1b[31m%s\x1b[0m', "Stopping connection process. No changes to existing config have been made.")
                    process.exit(1);
                }
            }

            //console.log("API key: ", apiKey)
            this.createCredentialsFromApiKey(apiKey)
        }

        //console.log('\x1b[35m%s\x1b[0m', "End of new Api key")
    
        return keyValid
    }
    
    checkApiKeyValidity = (apiKey) => {
        // Question: Is this all? It seems to be a very superficial check?

        //console.log('\x1b[35m%s\x1b[0m', "Validating Key")

        apiKey = apiKey.replace(/\s/g, "")

        const dots = (apiKey.match(/\./g) || []).length

        if (dots === 1){
            const [prefix, secretKey] = apiKey.split(".")
            if (prefix.length === 7 && secretKey.length === 33){
                //console.log("Key is valid")
                return true
            }
        }

        //console.log('\x1b[35m%s\x1b[0m', "Ending key validation")
        return false
    }

    createCredentialsFromApiKey = (apiKey) => {

        apiKey = apiKey.replace(/\s/g, "")
        const [prefix, secretKey] = apiKey.split(".")

        //console.log(apiKey)

        // Now two hashes are created. I don't know what the variable names
        // mean, they were like that in the old version. The hashes are then
        // used in the clientCredentials for the API 

        this.uniFunctions.logger.verbose("Generating credentials from API key")

        try {
            const m_one = crypto.createHash("sha256") 
            m_one.update(Buffer.from(apiKey, "utf-8").toString())
            m_one.update(Buffer.from(prefix, "utf-8").toString())
            
            const m_one_digest = m_one.digest()
            
            const m_one_steptwo = crypto.createHash("sha256")
            m_one_steptwo.update(m_one_digest)
            m_one_steptwo.update(Buffer.from(this.uniFunctions.serial, "utf-8").toString())
            
            const hash_c_one = m_one_steptwo.digest('hex')
    
            const n_two = crypto.createHash("sha256")
            n_two.update(Buffer.from(apiKey, "utf-8").toString())
            n_two.update(Buffer.from(apiKey, "utf-8").toString())
            n_two.update(Buffer.from(prefix, "utf-8").toString())
    
            const hash_c_two = n_two.digest("hex")
            
            this.finalizeCredentials(hash_c_one, hash_c_two, prefix)
            this.uniFunctions.logger.verbose("Finished generating credentials from key")
        } 
        catch (err) {
            this.uniFunctions.logger.error("ERROR: Error generating credentials from API key", err)
            // throw new Error("Error generating credentials from API key")
            throw {"ERROR": "Error generating credentials from API key"}
        }
    }

    finalizeCredentials = (hash_c_one, hash_c_two, prefix) => {
        // Encrypts the credentials previously generated with the api key

        this.uniFunctions.logger.verbose("Encrypting and saving credentials")

        const hr_one = this.uniFunctions.symmetricEncode("", hash_c_one)
        const hr_two = this.uniFunctions.symmetricEncode(hr_one["salt"], hash_c_two)
        const hr_prefix = this.uniFunctions.symmetricEncode(hr_one["salt"], prefix)

        this.uniFunctions.config["apikeys"][this.uniFunctions.encodedSerial] = {
            "hash_c_one": hr_one["token"],
            "hash_c_two": hr_two["token"],
            "prefix": hr_prefix["token"],
            "mqtt_password": "",
            "mqtt_salt": "",
            "cred_salt": hr_one["salt"]
        }

        this.uniFunctions.logger.verbose("Finished encrypting and saving credentials")

        this.uniFunctions.updateSettings()
    }

    // --- Pi setup --- 

    addPi = async (piName) => {

        this.uniFunctions.logger.info("Adding new pi")

        let name

        if(piName){
            name = piName
        }
        else{
            console.log('\x1b[35m%s\x1b[0m', "Enter the name for the pi: ")
            name = prompt("       ")
        }

        if (this.uniFunctions.decryptedClientCredentials && name){

            let eth0Mac, wlan0Mac, piModel
            
            try {
                eth0Mac = fs.existsSync("/sys/class/net/eth0/address") ?
                    fs.readFileSync("/sys/class/net/eth0/address").toString().replace("\n", "") : 
                    null
                wlan0Mac = fs.existsSync("/sys/class/net/wlan0/address") ?
                    fs.readFileSync("/sys/class/net/wlan0/address").toString().replace("\n", "") :
                    null
                piModel = fs.readFileSync("/proc/device-tree/model").toString().replace("\x00", "")
            }
            catch (err) {
                this.uniFunctions.logger.error("ERROR: Error reading pidata")
                // throw new Error("Error reading pidata")
                throw {"ERROR": "Error reading pidata"}
            }

            let hasWlan
            let hasLan

            wlan0Mac ? hasWlan = true : hasWlan = false
            eth0Mac ? hasLan = true : hasLan = false

            const pidata = {
                "serial": this.uniFunctions.serial,
                "revision": null,
                "model": piModel,
                "name": name,
                "lan_mac": eth0Mac,
                "wlan_mac": wlan0Mac,
                "has_lan": hasLan,
                "has_wlan": hasWlan
            }
            //console.log(pidata)


            const apiAnswer = await this.uniFunctions.api.addPi(pidata)

            if (apiAnswer){
                this.uniFunctions.logger.info("The pi was successfully added to picockpit.com")
            }
            else{
                this.uniFunctions.logger.error(`ERROR: The pi was not added to picockpit, API answer: ${apiAnswer}`)
                // throw new Error("The pi was not added to picockpit")
                throw {"ERROR": "The pi was not added to picockpit"}
            }

            //console.log('\x1b[35m%s\x1b[0m', "Ended adding pi")

            return apiAnswer
        }

    }

    // --- Mqtt setup ---

    setUpMqtt = async () => {
        const apiAnswer = await this.getMqttCredentials()
        this.testMqttConnection(apiAnswer)
        return (!apiAnswer.error)
    }

    getMqttCredentials = async () => {
        this.uniFunctions.logger.verbose("Getting new mqtt-credentials")
        
        const credentials = await this.uniFunctions.api.getMqttCredentials()
        
        this.uniFunctions.logger.verbose("Finished getting mqtt creds")
        return credentials
    }

    testMqttConnection = async (connectionData, tries=5) => {
        const { mqtt_server, mqtt_port, mqtt_path } = this.uniFunctions.config
        const { client_id, username, password } = connectionData

        //this.uniFunctions.verboseLog("Testing mqtt-connection")

        if (mqtt_port && mqtt_server){
            const mqttUri = `wss://${mqtt_server}:${mqtt_port}${mqtt_path}`
            const connectionOptions = {
                clientId: client_id,
                username,
                password
            }
            const mqttClient = mqtt.connect(
                mqttUri,
                connectionOptions
            )

            mqttClient.on("connect", () => {
                console.log('\x1b[32m%s\x1b[0m', "Connection successful")
                console.log('\x1b[32m%s\x1b[0m', "       The client will start automatically if you are using an install script.")
                console.log('\x1b[32m%s\x1b[0m', "       Otherwise you can either:")
                console.log('\x1b[32m%s\x1b[0m', "         1) Run the client manually with 'sudo picockpit-client daemon'.")
                console.log('\x1b[32m%s\x1b[0m', "         2\) Start the systemd-service with 'sudo systemctl start picockpit-client'.")
                console.log('\x1b[32m%s\x1b[0m', "            Enable auto-start with sytemd-service with 'sudo systemctl enable picockpit-client' afterwards.")

                mqttClient.end()
                this.saveMqttCredentials(connectionData)
            })

            mqttClient.on("error", (err) => {
                console.log("Error: ", err)
                mqttClient.end()
                if (tries > 0){
                    this.testMqttConnection(connectionData, tries-1)
                }
                else{
                    this.uniFunctions.logger.error("ERROR: Mqtt connection test failed")
                    // throw new Error("Mqtt connection-test failed")
                    throw {"ERROR": "Mqtt connection-test failed"}
                }
            })
        }

        //this.uniFunctions.verboseLog("Finished testing mqtt")
    }

    saveMqttCredentials = (connectionData) => {

        this.uniFunctions.logger.verbose("Encrypting and saving mqtt-credentials")

        const hexRet = this.uniFunctions.symmetricEncode("", connectionData["password"])
        const { token, salt } = hexRet

        this.uniFunctions.config["apikeys"][this.uniFunctions.encodedSerial]["mqtt_password"] = token
        this.uniFunctions.config["apikeys"][this.uniFunctions.encodedSerial]["mqtt_salt"]= salt
        
       this.uniFunctions.updateSettings() 

       this.uniFunctions.logger.verbose("Finished encrypting and saving mqtt-credentials")
    }
}
