import Fastify from 'fastify'
import { spawn } from 'child_process';
import fs from 'fs';
const fastify = Fastify();

let jobQueue = [];
let jobResults = [];
let runner_running = false;


// There is no error management in here, since the functioncall itself is wrapped in the queueRunner (Or however it will be called at some point).
// This is intentional - either a script runs (Good), a script has an internal error (Error returns and executeScriptAsync works) or there is an internal
// error to the function executeScriptAsync in which case we do want it to crash and be handeled by the runner instead of the function doing something weird
// and trying to error-handle internally.
function executeScriptAsync(scriptContent) {
  return new Promise((resolve, reject) => {
    const childProcess = spawn('bash', ['-c', scriptContent], {timeout : 600000});
    var output = '';

    // Listen for stdout data event
    childProcess.stdout.on('data', (data) => {
      output = output + data;
    });

    // Listen for stderr data event
    childProcess.stderr.on('data', (data) => {
      output = output + data;
    });
    
    // Listen for child process close event
    childProcess.on('exit', (code, signal) => {
    // Listen for child process timeout event
      if (signal === 'SIGTERM') {
        reject(new Error("Process timed out."));
    } else {
      resolve([output, code]);
    }
    });

    // Listen for child process error event
    childProcess.on('error', (err) => {
      reject([output, code]);
    });
  });
}

// runs async all jobs in queue
async function queueRunner() {
  if (runner_running) {return;}

  try {
    runner_running = true;
    while (jobQueue.length > 0) {
        const job = jobQueue[0]
        const id = job._id
        const body = job.body
        const ts = job.ts
        const name = job.name
        const c_name = job.c_name
        try {
          //console.log("Running: ", job);
          const code = await executeScriptAsync(body);
          jobResults.push({_id: id, timestamp: ts, c_name, name, output: code[0], retval: code[1]});
          jobQueue.shift()
        } catch (error) { 
          const val = `Error: ${error.message}`;
          console.error('XXXX Script execution (Not the script itself) encountered an error:', error);
          console.error('Script execution (Not the script itself) encountered an error:', val);
          jobResults.push({_id: id, timestamp: ts, c_name, name, output: val, retval: -1});
          jobQueue.shift()
          //console.error('Script execution (Not the script itself) encountered an error:', error);
        }
      }  
  } catch (error) {
    console.error("Runner crashed, probably while parsing the job Queue.", error);
    // We have to make sure that even if the runner crashes, we set runner_running to false so a new one can start up.
    runner_running = false;
    // We also reset the job queue and publish a 'catastrophic result' TODO.
    // In the future it is debatable whether the whole jbqueuemgr should restart, etc.
    // jobResults.push({_id: id, timestamp: ts, c_name, name, output: code[0], retval: code[1]});
    jobQueue = []
    return;
  }
  runner_running = false;
}

fastify.post('/submit-job', function (request, reply) {
  try {
    const requestBody = request.body.jobs;
    for (let i = 0; i < requestBody.length; i++) {
      const timestamp = Date.now();
      var job = { ...requestBody[i], c_id:request.body.c_id, c_name:request.body.c_name }
      job["ts"] = timestamp
      jobQueue.push(job)
    } 
  } catch (error) {
    reply.send({ statusCode: 501, message: 'Error while parsing the jobs and adding them to the execution queue.' });
  }
    try {
      queueRunner();
    } catch (error) {
      reply.send({ statusCode: 502, message: 'Error while starting the runner.' });
    }
  reply.send({ statusCode: 200, message: 'success' }); 
});

fastify.get('/poll-job-results', function (request, reply) {
  var jqLen;
  var jrLen;
  var results;
  try {
    jqLen = jobQueue.length;
    jrLen = jobResults.length;
    results = jobResults;
  } catch (err) {
    reply.send({ statusCode: 502, message: 'Error while parsing the completed jobs.' });
  }
  reply.send({statusCode: 200, jqLen:jqLen, jrLen:jrLen, results:results});
  jobResults = [];
})


// Path to the config file
const conf_file_Path = '/etc/picockpit-client/svc.d/jobqueuemgr.json';
var port;
var jsonData;

try {
  const data = fs.readFileSync(conf_file_Path, 'utf8');
  var jsonData = JSON.parse(data);
} catch {
  console.error('Error reading configuration file. \n it should be located at /etc/picockpit-client/svc.d/jobqueuemgr.json\n' + 
  'There is probably an issue with missing or corrputed configuration files.\n' + 
  'Try reinstalling Picockpit from a new .deb file, either via the Website or via the Repository.\nContact our team at www.picockpit.com if the error persists.', err);
  // We simply exit here since the systemd service takes care of restarting & testing again.
  process.exit(1)
}
try {
  port = jsonData["com_port"];
} catch {
  console.error('Error parsing configuration file. \n it should be located at /etc/picockpit-client/svc.d/jobqueuemgr.json\n' + 
  'There is probably an issue with missing or corrputed configuration files.\n' + 
  'Try reinstalling Picockpit from a new .deb file, either via the Website or via the Repository.\nContact our team at www.picockpit.com if the error persists.', err);
  // We simply exit here since the systemd service takes care of restarting & testing again.
  process.exit(1)
}

fastify.listen({ port:port }, function (err, address) {
  console.log("Starting up the jobqueuemgr on port: ", port)
  if (err) {
    console.error("There has been an error starting up or running the server for picockpit-internal communication.\n" +
                  "This is probably an issue with ports, the server starts up on the port defined inside /etc/picockpit-client/svc.d/jobqueuemgr.json\n" +
                  "You can double check if this port is already occupied on your system.\n" +
                  "Should this not be the case please write us on picockpit.com for further help.")
    // Note that this error doesn't include a message for the server - if the config file is missing we obviously can't reach the server.
    fastify.log.error(err)
    process.exit(1)
  }
})


function read_job_queue() {

}

function add_to_job_queue() {

}


function add_to_results_list(){

}

function read_results_list(){

}
