const mqtt = require('mqtt')
const fs = require('fs')
const { spawnSync } = require('child_process')
const winston = require('winston');
const { combine, timestamp, printf, colorize, align, json } = winston.format;
const path = require('path');

// --- Get raspi serial (Required for mqtt paths) ---

module.exports = class PCCApp {
    constructor(appName) {
        this.name = appName
        this.appPath = `../picockpit-apps/com.picockpit/${appName}`
        this.configDirectory = `/etc/picockpit-client/apps/com.picockpit/${this.name}`
        this.configPath = `/etc/picockpit-client/apps/com.picockpit/${this.name}/config.json`
        
        this.printLogLevel
        this.saveLogLevel

        this.readConfig()

        this.logger = winston.createLogger({
            levels: { silent: 0, error: 1, warn: 2, info: 3, http: 4, verbose: 5, debug: 6, silly: 7 },
            transports: [
                new winston.transports.Console({
                    level: this.printLogLevel || "error",
                    format: combine(
                        colorize({all: true}),
                        timestamp({
                            format: 'hh:mm:ss',
                        }),
                        align(),
                        printf((info) => `[${info.timestamp}][${appName}] ${info.level}: ${info.message}`)
                    ) 
                }),
                new winston.transports.File({
                    levels: { silent: 0, error: 1, warn: 2, info: 3, http: 4, verbose: 5, debug: 6, silly: 7 },
                    level: this.saveLogLevel || "silent",
                    filename: `/var/log/picockpit-client/${appName}.log`,
                    format: combine(
                        timestamp({
                          format: 'YYYY-MM-DD hh:mm:ss.SSS',
                        }),
                        json()
                    ),
                })
            ],
        });

        this.raspiSerial = this.getRaspiSerial()
        this.appSetup()

        this.setUpMqtt()
    }
    setUpMqtt = async () => {
        this.logger.verbose("Setting up mqtt-connection")

        this.prepareMqtt()
        this.connected = false

        this.client = mqtt.connect(this.mqttUri, this.connectionOptions)
        
        this.client.on('connect', () => {
            this.connected = true
            this.logger.verbose('Mqtt connection established')
        }),
        this.client.on('error', () => {
            this.connected = false
            this.logger.warn('Error connecting via mqtt')
        })

        this.subscribe()
    }
    subscribe = () => {
        this.logger.verbose('Subscribing to following topics')
        if (this.subscriptionTopics){
            for (const item of this.subscriptionTopics) {
                this.logger.verbose(`pi/${this.raspiSerial}/com.picockpit/${item}`)
                this.client.subscribe(`pi/${this.raspiSerial}/com.picockpit/${item}`)
            }
        }
    }
    publish = (topic, msg, options = {}) => {
        if (this.connected){
            this.logger.debug(`Publishing to: ${topic}`)
            // this.debugLog(`msg: ${JSON.stringify(util.inspect(msg, {showHidden: false, depth: null, colors: true}))}`)
            const {retain = false, raw = false, appName = this.name} = options
            if (!raw){
                msg = JSON.stringify(msg)
            }
            const completeTopic = `com.picockpit/${appName}/${topic}`
            const testTopic = `internal/${completeTopic}`
            this.client.publish(testTopic, msg, {retain})
        }
        else {
            this.logger.warn(`App trying to publish without connection, topic: ${topic}`)
        }
    }
    prepareMqtt = () => {
        this.connectionOptions = {
            clientId: this.name,
            username: "test",
            password: "testpassword",
            timeout: 30
        }
        
        this.mqttServer = "localhost"
        this.mqttPort = 1883
        this.mqttUri = `tcp://${this.mqttServer}:${this.mqttPort}`
    }
    readConfig = () => {
        const configDirectory = `/etc/picockpit-client/apps/com.picockpit/${this.name}`
        
        if (!fs.existsSync(configDirectory)){
            fs.mkdirSync(configDirectory)
        }
        if (fs.existsSync(this.configPath)){
            const rawConfig = fs.readFileSync(this.configPath)
            this.config = JSON.parse(rawConfig)
        }
        else if (fs.existsSync(this.defaultConfigPath)){
            const rawConfig = fs.readFileSync(this.defaultConfigPath)
            fs.copyFileSync(this.defaultConfigPath, this.configPath)
            this.config = JSON.parse(rawConfig)
        }
        else {
            // this.logger.error("ERROR: No default config file found")
            throw new Error('No default config file found')
        }

        this.subscriptionTopics = this.config["subscribe"]
        this.printLogLevel = this.config["printLogLevel"]
        this.saveLogLevel = this.config["saveLogLevel"]
        this.setupscriptExists = this.config["setupscript_exists"]
        this.setupscriptExecuted = this.config["setupscript_executed"] 
    }
    appSetup = () => {
        // TODO: Add Comments
        if (this.setupscriptExists && !this.setupscriptExecuted){
            fs.readdirSync(`${this.appPath}/setup`).forEach(file => {
                spawnSync(`${this.appPath}/setup/${file}`)
            });
            this.config["setupscript_executed"] = true
            fs.writeFileSync(this.configPath, JSON.stringify(this.config))
        }
    }
    getRaspiSerial = () => {
        let raspiSerial = fs.readFileSync('/sys/firmware/devicetree/base/serial-number', 'utf-8', (err) => {
            if (err) {
                this.logger.error("ERROR: Reading raspi serial number failed")
                new Error('Reading raspi serial number failed')
            }
        })
        raspiSerial = JSON.stringify(raspiSerial)
        raspiSerial = raspiSerial.slice(1, -7)
        return raspiSerial
    }
}
